/* ------------------------------------------------------------------------- */
/*
 *  socket.h
 *
 *  Copyright (c) 2004 - 2008, clown. All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *    - Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *    - Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in the
 *      documentation and/or other materials provided with the distribution.
 *    - No names of its contributors may be used to endorse or promote
 *      products derived from this software without specific prior written
 *      permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 *  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 *  OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *  Last-modified: Fri 16 Mar 2007 17:52:03 JST
 */
/* ------------------------------------------------------------------------- */
#ifndef CLX_SOCKET_H
#define CLX_SOCKET_H

#ifndef CLX_WIN32
#if defined(_WIN32) || defined(__WIN32__) || defined(WIN32)
#define CLX_WIN32
#endif
#endif // CLX_WIN32

#include <cerrno>
#include <cstdlib>
#include <cstring>
#include <string>
#include <sstream>
#include <stdexcept>

#ifdef CLX_WIN32
#include "winsock_util.h"
#include <windows.h>
#include <io.h>
#else
#include <sys/socket.h>
#include <sys/types.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <unistd.h>
#endif // CLX_WIN32

#ifdef CLX_WIN32
typedef SOCKET socket_int;
typedef int socklen_type;
#else
typedef int socket_int;
typedef socklen_t socklen_type;
#endif // CLX_WIN32

namespace clx {
	/* --------------------------------------------------------------------- */
	//  socket_error
	/* --------------------------------------------------------------------- */
	class socket_error : public std::runtime_error {
	public:
		typedef char char_type;
		typedef std::string string_type;
		typedef std::stringstream sstream_type;
		
		static const int nbuf = 1024;
		
		explicit socket_error(const string_type& what_arg) :
			std::runtime_error(what_arg) {
#ifdef CLX_WIN32
			errno_ = ::WSAGetLastError();
#else
			errno_ = errno;
#endif
		}
		
		virtual const char_type* what() {
			sstream_type tmp;
			char message[nbuf];
#ifdef CLX_WIN32
			strerror_s(message, sizeof(message), errno_);
#else
			strerror_r(errno_, message, sizeof(message));
#endif
			tmp << std::runtime_error::what() << ": " << message;
			return tmp.str().c_str();
		}
		
	private:
		int errno_;
	};
	
	/* --------------------------------------------------------------------- */
	//  basic_rawsocket
	/* --------------------------------------------------------------------- */
	template <
		int Type,
		int Family
	>
	class basic_rawsocket {
	public:
		// constructor and destructor
		basic_rawsocket() : s_(-1) {}
		
		basic_rawsocket(const basic_rawsocket& cp) : s_(-1) {
			this->assign(cp);
		}
		
		explicit basic_rawsocket(socket_int s) : s_(-1) {
			this->assign(s);
		}
		
		virtual ~basic_rawsocket() { this->close(); }
		
		basic_rawsocket& operator=(const basic_rawsocket& cp) {
			return this->assign(cp);
		}
		
		basic_rawsocket& operator=(socket_int s) {
			return this->assign(s);
		}
		
		// operator
		basic_rawsocket& assign(const basic_rawsocket& cp) {
			if (&cp == this) return *this;
			this->close();
#ifdef CLX_WIN32
			if (cp.s_ != -1) {
				::DuplicateHandle(::GetCurrentProcess(), (HANDLE)cp.s_,
					::GetCurrentProcess(), (LPHANDLE)&s_,
					0, TRUE, DUPLICATE_SAME_ACCESS);
			}
#else
			if (cp.s_ != -1) { s_ = ::dup(cp.s_); }
#endif
			return *this;
		}
		
		basic_rawsocket& assign(socket_int s) {
			this->close();
			s_ = s;
			return *this;
		}
		
		socket_int open() {
			this->close();
			s_ = ::socket(Family, Type, 0);
			return s_;
		}
		
		void close() {
			if (s_ >= 0) {
#ifdef CLX_WIN32
				::closesocket(s_);
#else
				::close(s_);
#endif
				s_ = -1;
			}
		}
		
		bool is_open() const { return s_ >= 0; }
		socket_int socket() const { return s_; }
		
	private:
		socket_int s_;
	};
}

#endif // CLX_SOCKET_H
